@@ -5,10 +5,12 @@
 //! ```rust,no_run
 //! extern crate postgres;
 //! extern crate postgres_binary_copy;
+//! extern crate streaming_iterator;
 //!
 //! use postgres::{Connection, TlsMode};
 //! use postgres::types::{Type, ToSql};
 //! use postgres_binary_copy::BinaryCopyReader;
+//! use streaming_iterator::StreamingIterator;
 //!
 //! fn main() {
 //!     let conn = Connection::connect("postgres://postgres@localhost",
@@ -20,17 +22,19 @@
 //!     let types = &[Type::Int4, Type::Varchar];
 //!     let data: Vec<Box<ToSql>> = vec![Box::new(1i32), Box::new("hello"),
 //!                                      Box::new(2i32), Box::new("world")];
-//!     let data = data.iter().map(|v| &**v);
+//!     let data = streaming_iterator::convert(data.into_iter()).map_ref(|v| &**v);
 //!     let mut reader = BinaryCopyReader::new(types, data);
 //!
 //!     let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN (FORMAT binary)").unwrap();
 //!     stmt.copy_in(&[], &mut reader).unwrap();
 //! }
 //! ```
-#![doc(html_root_url="https://sfackler.github.io/rust-postgres-binary-copy/doc/v0.3.1")]
+#![doc(html_root_url="https://docs.rs/postgres-binary-copy/0.4.0")]
 #![warn(missing_docs)]
+
 extern crate byteorder;
 extern crate postgres;
+extern crate streaming_iterator;
 
 use byteorder::{BigEndian, ReadBytesExt, WriteBytesExt};
 use postgres::types::{Type, ToSql, IsNull};
@@ -40,34 +44,10 @@ use std::fmt;
 use std::io::prelude::*;
 use std::io::{self, Cursor};
 use std::mem;
+use streaming_iterator::StreamingIterator;
 
 const HEADER_MAGIC: &'static [u8] = b"PGCOPY\n\xff\r\n\0";
 
-/// Like `Iterator`, except that it returns borrowed values.
-///
-/// In contrast to `Iterator<Item = &T>`, a type implementing
-/// `StreamingIterator<Item = T>` does not need to have all of the values it
-/// returns in memory at the same time.
-///
-/// All `Iterator`s over `&T` are also `StreamingIterator`s over `T`.
-pub trait StreamingIterator {
-    /// The type of elements being iterated.
-    type Item: ?Sized;
-
-    /// Advances the iterator and returns the next value.
-    ///
-    /// Returns `None` when the end is reached.
-    fn next(&mut self) -> Option<&Self::Item>;
-}
-
-impl<'a, T: 'a + ?Sized, I: Iterator<Item = &'a T>> StreamingIterator for I {
-    type Item = T;
-
-    fn next(&mut self) -> Option<&T> {
-        unsafe { std::mem::transmute(Iterator::next(self)) }
-    }
-}
-
 #[derive(Debug, Copy, Clone)]
 enum ReadState {
     Header,
@@ -89,10 +69,10 @@ impl<'a, I> fmt::Debug for BinaryCopyReader<'a, I>
 {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
         fmt.debug_struct("BinaryCopyReader")
-           .field("types", &self.types)
-           .field("state", &self.state)
-           .field("it", &self.it)
-           .finish()
+            .field("types", &self.types)
+            .field("state", &self.state)
+            .field("it", &self.it)
+            .finish()
     }
 }
 
@@ -117,7 +97,7 @@ impl<'a, I> BinaryCopyReader<'a, I>
         }
     }
 
-    fn fill_buf(&mut self, info: &CopyInfo) -> io::Result<()> {
+    fn fill_buf(&mut self, _: &CopyInfo) -> io::Result<()> {
         enum Op<'a> {
             Value(usize, &'a ToSql),
             Footer,
@@ -134,7 +114,8 @@ impl<'a, I> BinaryCopyReader<'a, I>
                 self.state = ReadState::Body(idx);
                 Op::Value(idx, value)
             }
-            (ReadState::Header, None) | (ReadState::Body(_), None) => {
+            (ReadState::Header, None) |
+            (ReadState::Body(_), None) => {
                 self.state = ReadState::Footer;
                 Op::Footer
             }
@@ -159,9 +140,7 @@ impl<'a, I> BinaryCopyReader<'a, I>
 
                 let len_pos = self.buf.position();
                 let _ = self.buf.write_i32::<BigEndian>(0); // space for length
-                let len = match value.to_sql_checked(&self.types[idx],
-                                                     self.buf.get_mut(),
-                                                     &info.session_info()) {
+                let len = match value.to_sql_checked(&self.types[idx], self.buf.get_mut()) {
                     Ok(IsNull::Yes) => -1,
                     Ok(IsNull::No) => {
                         let len = self.buf.get_ref().len() as u64 - 4 - len_pos;
@@ -193,7 +172,7 @@ impl<'a, I> ReadWithInfo for BinaryCopyReader<'a, I>
 {
     fn read_with_info(&mut self, buf: &mut [u8], info: &CopyInfo) -> io::Result<usize> {
         if self.buf.position() == self.buf.get_ref().len() as u64 {
-            try!(self.fill_buf(info));
+            self.fill_buf(info)?;
         }
         self.buf.read(buf)
     }
@@ -228,10 +207,7 @@ enum WriteState {
     AtHeader,
     AtTuple,
     AtFieldSize(usize),
-    AtField {
-        size: usize,
-        remaining: usize,
-    },
+    AtField { size: usize, remaining: usize },
     Done,
 }
 
@@ -249,11 +225,11 @@ impl<W> fmt::Debug for BinaryCopyWriter<W>
 {
     fn fmt(&self, fmt: &mut fmt::Formatter) -> fmt::Result {
         fmt.debug_struct("BinaryCopyWriter")
-           .field("state", &self.state)
-           .field("has_oids", &self.has_oids)
-           .field("value_writer", &self.value_writer)
-           .field("buf", &self.buf.len())
-           .finish()
+            .field("state", &self.state)
+            .field("has_oids", &self.has_oids)
+            .field("value_writer", &self.value_writer)
+            .field("buf", &self.buf.len())
+            .finish()
     }
 }
 
@@ -274,28 +250,28 @@ impl<W> BinaryCopyWriter<W>
 
     fn read_to(&mut self, buf: &[u8], size: usize) -> io::Result<(bool, usize)> {
         let to_read = cmp::min(size - self.buf.len(), buf.len());
-        let nread = try!(self.buf.write(&buf[..to_read]));
+        let nread = self.buf.write(&buf[..to_read])?;
         Ok((nread == to_read, nread))
     }
 
     fn read_header(&mut self, buf: &[u8]) -> io::Result<usize> {
         let header_size = HEADER_MAGIC.len() + mem::size_of::<i32>() * 2;
-        let (done, nread) = try!(self.read_to(buf, header_size));
+        let (done, nread) = self.read_to(buf, header_size)?;
         if !done {
             return Ok(nread);
         }
 
-        if &self.buf[..HEADER_MAGIC.len()] != HEADER_MAGIC {
+        if !self.buf.starts_with(HEADER_MAGIC) {
             return Err(io::Error::new(io::ErrorKind::InvalidInput, "invalid header"));
         }
 
-        let flags = try!((&mut &self.buf[HEADER_MAGIC.len()..]).read_i32::<BigEndian>());
+        let flags = (&mut &self.buf[HEADER_MAGIC.len()..])
+            .read_i32::<BigEndian>()?;
 
         self.has_oids = (flags & 1 << 16) != 0;
 
         if (flags & !0 << 17) != 0 {
-            return Err(io::Error::new(io::ErrorKind::InvalidInput,
-                                      "critical file format issue"));
+            return Err(io::Error::new(io::ErrorKind::InvalidInput, "critical file format issue"));
         }
 
         self.buf.clear();
@@ -304,12 +280,12 @@ impl<W> BinaryCopyWriter<W>
     }
 
     fn read_tuple(&mut self, buf: &[u8]) -> io::Result<usize> {
-        let (done, nread) = try!(self.read_to(buf, mem::size_of::<i16>()));
+        let (done, nread) = self.read_to(buf, mem::size_of::<i16>())?;
         if !done {
             return Ok(nread);
         }
 
-        let mut tuple_size = try!((&mut &self.buf[..]).read_i16::<BigEndian>());
+        let mut tuple_size = (&mut &self.buf[..]).read_i16::<BigEndian>()?;
 
         self.buf.clear();
         if tuple_size == -1 {
@@ -329,16 +305,16 @@ impl<W> BinaryCopyWriter<W>
                        info: &CopyInfo,
                        remaining: usize)
                        -> io::Result<usize> {
-        let (done, nread) = try!(self.read_to(buf, mem::size_of::<i32>()));
+        let (done, nread) = self.read_to(buf, mem::size_of::<i32>())?;
         if !done {
             return Ok(nread);
         }
 
-        let field_size = try!((&mut &self.buf[..]).read_i32::<BigEndian>());
+        let field_size = (&mut &self.buf[..]).read_i32::<BigEndian>()?;
 
         self.buf.clear();
         if field_size == -1 {
-            try!(self.value_writer.write_null_value(info));
+            self.value_writer.write_null_value(info)?;
             self.advance_field_state(remaining);
         } else {
             self.state = WriteState::AtField {
@@ -363,12 +339,12 @@ impl<W> BinaryCopyWriter<W>
                   size: usize,
                   remaining: usize)
                   -> io::Result<usize> {
-        let (done, nread) = try!(self.read_to(buf, size));
+        let (done, nread) = self.read_to(buf, size)?;
         if !done {
             return Ok(nread);
         }
 
-        try!(self.value_writer.write_value(&self.buf, info));
+        self.value_writer.write_value(&self.buf, info)?;
         self.buf.clear();
         self.advance_field_state(remaining);
         Ok(nread)
@@ -397,6 +373,7 @@ mod test {
     use postgres::{Connection, TlsMode};
     use postgres::types::{Type, FromSql, ToSql};
     use postgres::stmt::CopyInfo;
+    use streaming_iterator::{convert, StreamingIterator};
 
     #[test]
     fn write_basic() {
@@ -405,19 +382,21 @@ mod test {
                      &[])
             .unwrap();
 
-        let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN BINARY").unwrap();
+        let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN BINARY")
+            .unwrap();
 
         let types = &[Type::Int4, Type::Varchar];
         let values: Vec<Box<ToSql>> = vec![Box::new(1i32),
                                            Box::new("foobar"),
                                            Box::new(2i32),
                                            Box::new(None::<String>)];
-        let values = values.iter().map(|e| &**e);
+        let values = convert(values.into_iter()).map_ref(|v| &**v);
         let mut reader = BinaryCopyReader::new(types, values);
 
         stmt.copy_in(&[], &mut reader).unwrap();
 
-        let stmt = conn.prepare("SELECT id, bar FROM foo ORDER BY id").unwrap();
+        let stmt = conn.prepare("SELECT id, bar FROM foo ORDER BY id")
+            .unwrap();
         assert_eq!(vec![(1i32, Some("foobar".to_string())), (2i32, None)],
                    stmt.query(&[])
                        .unwrap()
@@ -433,7 +412,8 @@ mod test {
                      &[])
             .unwrap();
 
-        let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN BINARY").unwrap();
+        let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN BINARY")
+            .unwrap();
 
         let types = &[Type::Int4, Type::Varchar];
         let mut values: Vec<Box<ToSql>> = vec![];
@@ -442,12 +422,13 @@ mod test {
             values.push(Box::new(format!("the value for {}", i)));
         }
 
-        let values = values.iter().map(|e| &**e);
+        let values = convert(values.into_iter()).map_ref(|v| &**v);
         let mut reader = BinaryCopyReader::new(types, values);
 
         stmt.copy_in(&[], &mut reader).unwrap();
 
-        let stmt = conn.prepare("SELECT id, bar FROM foo ORDER BY id").unwrap();
+        let stmt = conn.prepare("SELECT id, bar FROM foo ORDER BY id")
+            .unwrap();
         let result = stmt.query(&[]).unwrap();
         assert_eq!(10000, result.len());
         for (i, row) in result.into_iter().enumerate() {
@@ -463,7 +444,8 @@ mod test {
                      &[])
             .unwrap();
 
-        let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN BINARY").unwrap();
+        let stmt = conn.prepare("COPY foo (id, bar) FROM STDIN BINARY")
+            .unwrap();
 
         let types = &[Type::Int4, Type::Bytea];
         let mut values: Vec<Box<ToSql>> = vec![];
@@ -472,12 +454,13 @@ mod test {
             values.push(Box::new(vec![i as u8; 128 * 1024]));
         }
 
-        let values = values.iter().map(|e| &**e);
+        let values = convert(values.into_iter()).map_ref(|v| &**v);
         let mut reader = BinaryCopyReader::new(types, values);
 
         stmt.copy_in(&[], &mut reader).unwrap();
 
-        let stmt = conn.prepare("SELECT id, bar FROM foo ORDER BY id").unwrap();
+        let stmt = conn.prepare("SELECT id, bar FROM foo ORDER BY id")
+            .unwrap();
         let result = stmt.query(&[]).unwrap();
         assert_eq!(2, result.len());
         for (i, row) in result.into_iter().enumerate() {
@@ -492,21 +475,16 @@ mod test {
         conn.execute("CREATE TEMPORARY TABLE foo (id SERIAL PRIMARY KEY, bar INT)",
                      &[])
             .unwrap();
-        conn.execute("INSERT INTO foo (bar) VALUES (1), (2), (NULL), (4)", &[]).unwrap();
+        conn.execute("INSERT INTO foo (bar) VALUES (1), (2), (NULL), (4)", &[])
+            .unwrap();
 
         let mut out = vec![];
 
         {
-            let writer = |r: Option<&[u8]>, info: &CopyInfo| {
+            let writer = |r: Option<&[u8]>, _: &CopyInfo| {
                 match r {
-                    Some(r) => {
-                        out.push(Option::<i32>::from_sql(&Type::Int4, r, &info.session_info())
-                                     .unwrap())
-                    }
-                    None => {
-                        out.push(Option::<i32>::from_sql_null(&Type::Int4, &info.session_info())
-                                     .unwrap())
-                    }
+                    Some(r) => out.push(Option::<i32>::from_sql(&Type::Int4, r).unwrap()),
+                    None => out.push(Option::<i32>::from_sql_null(&Type::Int4).unwrap()),
                 }
                 Ok(())
             };
@@ -514,7 +492,7 @@ mod test {
             let mut writer = BinaryCopyWriter::new(writer);
 
             let stmt = conn.prepare("COPY (SELECT bar FROM foo ORDER BY id) TO STDOUT BINARY")
-                           .unwrap();
+                .unwrap();
             stmt.copy_out(&[], &mut writer).unwrap();
         }
 
@@ -524,7 +502,8 @@ mod test {
     #[test]
     fn read_many_rows() {
         let conn = Connection::connect("postgres://postgres@localhost", TlsMode::None).unwrap();
-        conn.execute("CREATE TEMPORARY TABLE foo (id INT)", &[]).unwrap();
+        conn.execute("CREATE TEMPORARY TABLE foo (id INT)", &[])
+            .unwrap();
 
         let mut expected = vec![];
         let stmt = conn.prepare("INSERT INTO foo (id) VALUES ($1)").unwrap();
@@ -536,15 +515,15 @@ mod test {
         let mut out = vec![];
 
         {
-            let writer = |r: Option<&[u8]>, info: &CopyInfo| {
-                out.push(i32::from_sql(&Type::Int4, r.unwrap(), &info.session_info()).unwrap());
+            let writer = |r: Option<&[u8]>, _: &CopyInfo| {
+                out.push(i32::from_sql(&Type::Int4, r.unwrap()).unwrap());
                 Ok(())
             };
 
             let mut writer = BinaryCopyWriter::new(writer);
 
             let stmt = conn.prepare("COPY (SELECT id FROM foo ORDER BY id) TO STDOUT BINARY")
-                           .unwrap();
+                .unwrap();
             stmt.copy_out(&[], &mut writer).unwrap();
         }
 
@@ -559,7 +538,8 @@ mod test {
             .unwrap();
 
         let mut expected = vec![];
-        let stmt = conn.prepare("INSERT INTO foo (id, bar) VALUES ($1, $2)").unwrap();
+        let stmt = conn.prepare("INSERT INTO foo (id, bar) VALUES ($1, $2)")
+            .unwrap();
         for i in 0..2i32 {
             let value = vec![i as u8; 128 * 1024];
             stmt.execute(&[&i, &value]).unwrap();
@@ -569,9 +549,8 @@ mod test {
         let mut out = vec![];
 
         {
-            let writer = |r: Option<&[u8]>, info: &CopyInfo| {
-                out.push(Vec::<u8>::from_sql(&Type::Bytea, r.unwrap(), &info.session_info())
-                             .unwrap());
+            let writer = |r: Option<&[u8]>, _: &CopyInfo| {
+                out.push(Vec::<u8>::from_sql(&Type::Bytea, r.unwrap()).unwrap());
                 Ok(())
             };
 
@@ -579,7 +558,7 @@ mod test {
 
             let stmt = conn.prepare("COPY (SELECT bar FROM foo ORDER BY id) TO STDOUT (FORMAT \
                                      binary)")
-                           .unwrap();
+                .unwrap();
             stmt.copy_out(&[], &mut writer).unwrap();
         }
 
@@ -589,26 +568,28 @@ mod test {
     #[test]
     fn read_with_oids() {
         let conn = Connection::connect("postgres://postgres@localhost", TlsMode::None).unwrap();
-        conn.execute("CREATE TEMPORARY TABLE foo (id INT) WITH OIDS", &[]).unwrap();
-        conn.execute("INSERT INTO foo (id) VALUES (1), (2), (3), (4)", &[]).unwrap();
+        conn.execute("CREATE TEMPORARY TABLE foo (id INT) WITH OIDS", &[])
+            .unwrap();
+        conn.execute("INSERT INTO foo (id) VALUES (1), (2), (3), (4)", &[])
+            .unwrap();
 
         let mut oids = vec![];
         let mut out = vec![];
 
         {
-            let writer = |r: Option<&[u8]>, info: &CopyInfo| {
+            let writer = |r: Option<&[u8]>, _: &CopyInfo| {
                 if oids.len() > out.len() {
-                    out.push(i32::from_sql(&Type::Bytea, r.unwrap(), &info.session_info())
-                                 .unwrap());
+                    out.push(i32::from_sql(&Type::Bytea, r.unwrap()).unwrap());
                 } else {
-                    oids.push(u32::from_sql(&Type::Oid, r.unwrap(), &info.session_info()).unwrap());
+                    oids.push(u32::from_sql(&Type::Oid, r.unwrap()).unwrap());
                 }
                 Ok(())
             };
 
             let mut writer = BinaryCopyWriter::new(writer);
 
-            let stmt = conn.prepare("COPY foo (id) TO STDOUT (FORMAT binary, OIDS)").unwrap();
+            let stmt = conn.prepare("COPY foo (id) TO STDOUT (FORMAT binary, OIDS)")
+                .unwrap();
             stmt.copy_out(&[], &mut writer).unwrap();
         }
 
